from __future__ import annotations

"""
The Base Extensible Client Class
"""

import abc
import contextlib
from lzl.logging import logger, null_logger, Logger
from lzl.types import eproperty
from typing import Optional, Type, TypeVar, Union, Set, List, Any, Dict, Literal, TYPE_CHECKING

if TYPE_CHECKING:
    from kvdb import KVDBSession
    from lzl.io import PersistentDict
    from lzl.pool import ThreadPool
    from lzo.registry.clients import RClientT
    from lzo.types.base import RegisteredSettings, BaseModel

    DataSchema = TypeVar('DataSchema', bound = BaseModel)
    ResultSchema = TypeVar('ResultSchema', bound = BaseModel)



class BaseClient(abc.ABC):
    """
    The Base Client Class
    """

    name: Optional[str] = None

    verbosity: Optional[int] = 0
    enable_registration: Optional[bool] = True
    schemas: Dict[str, Union[Type['DataSchema'], Type['ResultSchema'], str]] = {}

    cache_enabled: Optional[bool] = True
    cache_expiration: Optional[int] = 60 * 60 * 24 * 3 # 3 days
    cache_shared_global: Optional[bool] = False
    cachable_default: Optional[bool] = None # The default value for cache enabled
    
    serialization: Optional[str] = 'json'
    serialization_compression: Optional[str] = 'zstd'
    serialization_compression_level: Optional[int] = 19 # 6 for default (lz4, zlib) 19 for zstd

    data_expiration: Optional[int] = None
    data_shared_global: Optional[bool] = False
    
    _extra: Dict[str, Any] = {}

    def __init__(self, *args, **kwargs):
        """
        Initializes the BaseClient
        """
        self._kwargs = kwargs
        self.pre_init(**kwargs)
        self.cls_init(**kwargs)
        self.post_init(**kwargs)
        self.configure_schemas(**kwargs)
        self.finalize_init(**kwargs)
        if self.enable_registration:
            self._register_client(**kwargs)
    

    def pre_init(self, **kwargs):
        """
        Pre-Initializes the BaseClient
        """
        pass

    def post_init(self, **kwargs):
        """
        Post-Initializes the BaseClient
        """
        pass

    def configure_schemas(self, **kwargs):
        """
        Configures the schemas
        """
        for name, schema in self.schemas.items():
            if isinstance(schema, str):
                schema = self._import(schema)
            self.schemas[name] = schema
        return self

    def cls_init(self, **kwargs):
        """
        Initializes the BaseClient
        """
        pass
    
    def finalize_init(self, **kwargs):
        """
        Finalizes the BaseClient
        """
        pass


    @eproperty
    def logger(self) -> 'Logger':
        """
        Gets the logger
        """
        return logger
    
    @eproperty
    def null_logger(self) -> 'Logger':
        """
        Gets the null logger
        """
        return null_logger
    
    @property
    def autologger(self) -> 'Logger':
        """
        Gets the auto logger
        """
        return self.logger if (
            self.settings.is_development_env or 
            self.verbosity >= 2
        ) else self.null_logger
    
    @eproperty
    def settings(self) -> 'RegisteredSettings':
        """
        Returns the settings
        """
        from lzo.registry.settings import get_app_settings
        return get_app_settings(self.__module__.__name__)
    
    @eproperty
    def pooler(self) -> 'ThreadPool':
        """
        Returns the thread pool
        """
        from lzl.pool import ThreadPool
        return ThreadPool
    

    @eproperty
    def kdb(self) -> 'KVDBSession':
        """
        [JSON Serializer] Returns the KVDB Session
        """
        return self.settings.ctx.get_kdb(self.name)
    
    @eproperty
    def rkdb(self) -> 'KVDBSession':
        """
        [Raw] Returns the Raw KVDB Session
        """
        return self.settings.ctx.get_kdb(f'{self.name}-raw', serializer = None, decode_responses = True)


    def get_extra_ser_pdict_kwargs(self, name: str, **kwargs) -> Dict[str, Any]:
        """
        Gets the extra serialization kwargs for the persistent dict
        """
        return {}

    @eproperty
    def data(self) -> 'PersistentDict':
        """
        Gets the Persistent Dict for data
        """
        serializer_kwargs = {
            'compression': self.serialization_compression,
            'compression_level': self.serialization_compression_level,
            'raise_errors': True,
            **self.get_extra_ser_pdict_kwargs('data'),
        }
        base_key = f'{self.settings.ctx.module_name}.{self.name}.data' if \
            self.data_shared_global else \
            f'{self.settings.ctx.module_name}.{self.settings.app_env.name}.{self.name}.data'
        return self.settings.ctx.get_pdict(
            base_key = base_key,
            expiration = self.data_expiration,
            serializer = self.serialization,
            serializer_kwargs = serializer_kwargs,
        )

    @eproperty
    def cache(self) -> 'PersistentDict[str, ResultSchema]':
        """
        Gets the Persistent Dict for cache
        """
        serializer_kwargs = {
            'compression': self.serialization_compression,
            'compression_level': self.serialization_compression_level,
            'raise_errors': True,
            **self.get_extra_ser_pdict_kwargs('cache'),
        }
        base_key = f'{self.settings.ctx.module_name}.{self.name}.cache' if \
            self.cache_shared_global else \
            f'{self.settings.ctx.module_name}.{self.settings.app_env.name}.{self.name}.cache'
        return self.settings.ctx.get_pdict(
            base_key = base_key,
            expiration = self.cache_expiration,
            serializer = self.serialization,
            serializer_kwargs = serializer_kwargs,
        )



    """
    Utilities
    """

    @contextlib.contextmanager
    def catch(self, exc: Type['Exception'], msg: Optional[str] = None, fatal: Optional[bool] = False):
        """
        Catches an exception
        """
        try:
            yield
        except Exception as e:
            log = self.logger.trace if fatal else self.autologger.trace
            log(f'Handling Error: {msg}', e, prefix = self.name)
            if fatal: raise exc(msg, error = e) from e



    def run(
        self,
        data: 'DataSchema',
        cachable: Optional[bool] = None,
        **kwargs,
    ) -> 'ResultSchema':
        """
        Runs the client
        """
        raise NotImplementedError("run must be implemented by the Client")
    
    async def arun(
        self,
        data: 'DataSchema',
        cachable: Optional[bool] = None,
        **kwargs,
    ) -> 'ResultSchema':
        """
        Runs the client
        """
        raise NotImplementedError("arun must be implemented by the Client")
    

    def __call__(
        self,
        data: 'DataSchema',
        cachable: Optional[bool] = None,
        is_async: Optional[bool] = True,
        **kwargs,
    ) -> 'ResultSchema':
        """
        Runs the Client
        """
        if cachable is None: cachable = self.cachable_default
        method = self.arun if is_async else self.run
        return method(data, cachable = cachable, **kwargs)


    """
    Helper Methods
    """
    
    def _import(self, name: str, **kwargs) -> Any:
        """
        Imports a module
        """
        from lzl.load import lazy_import
        return lazy_import(name, **kwargs)
    

    def _register_client(self, **kwargs):
        """
        Registers the client
        """
        self.settings.ctx.register_client(self, **kwargs)


    def _get_client(self, name: str, **kwargs) -> 'RClientT':
        """
        Gets a client
        """
        return self.settings.ctx.get_client(name, **kwargs)


from __future__ import annotations

"""
Hatchet Session with Modifications
"""

import asyncio
from hatchet_sdk.rate_limit import RateLimit
from hatchet_sdk.client import ClientImpl

from hatchet_sdk.hatchet import Hatchet as BaseHatchet
from hatchet_sdk.clients.rest.models.workflow_runs_cancel_request import (
    WorkflowRunsCancelRequest,
)
from hatchet_sdk.clients.rest.models import WorkflowRunStatus
from hatchet_sdk.clients.run_event_listener import StepRunEventType
from lzl.logging import logger
from kvdb.utils.cron import validate_cron_schedule
from ..utils import new_session, get_hatchet_settings, set_hatchet_settings, set_hatchet_client
from .version import CURRENT_VERSION as VERSION
from typing import Any, Dict, List, Optional, Union, Callable, TYPE_CHECKING, ParamSpec, Type, TypeVar, Tuple, overload

if TYPE_CHECKING:
    from lzo.utils.logs import Logger
    from lzl.api.hatchet.config import HatchetSettings
    from hatchet_sdk.clients.rest.models import WorkflowRun
    from hatchet_sdk.loader import ClientConfig
    from ..workflow import WorkflowT
    from .worker import Worker
    from ..state import GlobalHatchetContext, GlobalHatchetContextObject
    from hatchet_sdk.workflows_pb2 import StickyStrategy


RT = TypeVar('RT')
P = ParamSpec("P")

class HatchetSession(BaseHatchet):
    client: ClientImpl
    name: str = 'hatchet_session'
    kind: str = 'client'

    def __init__(
        self, 
        debug: bool = False, 
        config: Union['ClientConfig', Dict[str, Any]] = None,
        settings: Optional['HatchetSettings'] = None,
        instance: Optional[str] = 'default',
        namespace: Optional[str] = None,
        api_endpoint: Optional[str] = None,
        grpc_endpoint: Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the Hatchet Session
        """
        ...

    def __init__(
        self, 
        debug: bool = True, 
        config: Union[ClientConfig, Dict[str, Any]] = None,
        settings: Optional['HatchetSettings'] = None,
        instance: Optional[str] = None,
        _ctx: Optional['GlobalHatchetContextObject'] = None,
        **kwargs,
    ):
        # We do not check whether the hatchet session is already initialized.
        # initialize the session
        if instance is None: instance = 'default'
        self.instance = instance
        if _ctx is None: 
            from ..state import GlobalHatchetContext
            _ctx = GlobalHatchetContext
        self.c = _ctx
        if settings is None: settings = get_hatchet_settings()
        else: set_hatchet_settings(settings)

        self.hsettings = settings
        # self._event_loop: Optional[asyncio.AbstractEventLoop] = None
        # self.ensure_loop()
        self.client = new_session(
            defaults = config, 
            settings = self.hsettings, 
            instance = self.instance, 
            **kwargs
        )
        self.has_namespace = self.client.config.namespace is not None
        # if not debug: logger.disable("hatchet_sdk")
        # else: self.hsettings.debug_enabled = True

        self.workers: Dict[str, 'Worker'] = {}
        self.registered_workflows: Dict[str, 'WorkflowT'] = {}
        self.bound_workflows: Dict[str, 'WorkflowT'] = {}
        self.compiled_workflows: Dict[str, Type['WorkflowT']] = {}
        
        self.post_init()
        self.finalize_init()
        # set_hatchet_client(self)

    # @property
    # def event_loop(self) -> asyncio.AbstractEventLoop:
    #     """
    #     Returns the event loop
    #     """
    #     if self._event_loop is None:
    #         from .spawn import get_event_loop
    #         self._event_loop = get_event_loop()
    #     return self._event_loop
    
    # def ensure_loop(self):
    #     """
    #     Ensures the event loop
    #     """
    #     if self._event_loop is None:
    #         from .spawn import get_event_loop
    #         self._event_loop = get_event_loop()

    @property
    def logger(self) -> 'Logger':
        """
        Gets the logger
        """
        return self.hsettings.logger
    
    @property
    def autologger(self) -> 'Logger':
        """
        Gets the autologger
        """
        return self.hsettings.autologger

    def post_init(self):
        """
        Hook for post init
        """
        pass

    def finalize_init(self):
        """
        Finalize the init
        """
        pass

    
    def _get_workflow(self, op_name: str, version: Optional[str] = None) -> Type['WorkflowT']:
        """
        Gets the workflow
        """
        raise NotImplementedError
    

    def get_worker(self, name: str, max_runs: int | None = None) -> 'Worker':
        """
        Creates a new worker
        """
        if self.hsettings.in_k8s:
            name += f'-{self.hsettings.host_name[-1]}'
        if name in self.workers: return self.workers[name]
        if max_runs is None: max_runs = self.hsettings.workflow_concurrency_limit
        worker = self.c.worker_class(
            name = name, 
            max_runs=max_runs, 
            config = self.client.config,
            session = self,
            context_class = self.c.context_class,
        )
        self.workers[name] = worker
        return worker
    

    def bind_step(
        self,
        obj: 'WorkflowT',
        func: Callable[P, RT],
        name: str = "",
        timeout: str = "60m",
        parents: List[str] = [],
        retries: int = 0,
        rate_limits: List[RateLimit] | None = None,
    ):  # sourcery skip: default-mutable-arg
        """
        Binds a step to a workflow
        """
        obj.bind(
            self.step(
                name = name,
                timeout = timeout,
                parents = parents,
                retries = retries,
                rate_limits = rate_limits,
                # trigger_name = obj.trigger_name,
                # obj = obj,
            )(func)
        )

    def bind_failure_step(
        self,
        obj: 'WorkflowT',
        func: Callable[P, RT],
        name: str = "",
        timeout: str = "",
        retries: int = 0,
        rate_limits: List[RateLimit] | None = None,
    ):
        """
        Binds a step to a workflow
        """
        obj.bind(
            self.on_failure_step(
                name = name,
                timeout = timeout,
                retries = retries,
                rate_limits = rate_limits,
            )(func)
        )

    def build_workflow(
        self,
        obj: Type['WorkflowT'],
        name: str = "",
        on_events: List[str] = [],
        on_crons: list = [],
        version: str = "",
        timeout: str = "60m",
        schedule_timeout: str = "15m",
        sticky: 'StickyStrategy' = None,
    ) -> Type['WorkflowT']:  # sourcery skip: default-mutable-arg
        """
        Binds the workflow component to the workflow
        """
        if obj._hatchet_patched: return obj
        obj.workflow_name = obj.op_name
        if self.hsettings.include_instance_name_in_workflow:
            obj.workflow_name = f'{self.instance}.{obj.workflow_name}'
        
        if obj.version: obj.workflow_name += f'.{obj.version}'
        if on_events:
            for n, event in enumerate(on_events):
                if obj.workflow_name not in event: on_events[n] = f'{obj.workflow_name}:{event}'
        else:
            on_events = [obj.op_name]

        if on_crons:
            for n, cron in enumerate(on_crons):
                on_crons[n] = validate_cron_schedule(cron)

        obj.on_events = on_events
        obj.on_crons = on_crons
        obj.name = name or obj.workflow_name
        obj.trigger_name = f'{self.client.config.namespace}{obj.name}' if self.has_namespace else obj.name
        obj.client = self.client
        obj.version = version or self.hsettings.version
        obj.timeout = timeout
        obj.schedule_timeout = schedule_timeout
        obj.sticky = sticky
        if getattr(obj, 'workflow_concurrency', None) and not hasattr(obj.concurrency, '_concurrency_fn_name'):
            obj.bind(
                self.concurrency(
                    # name = f'{obj.op_name}:concurrency',
                    max_runs = obj.workflow_concurrency,
                    limit_strategy = obj.workflow_concurrency_strategy,
                )(obj.concurrency))
        if hasattr(obj, 'register_workflow_steps'):
            obj.register_workflow_steps(self)
        obj._hatchet_patched = True
        self.registered_workflows[obj.name] = obj
        return obj


    def bind_workflow(
        self,
        obj: 'WorkflowT',
        name: str = "",
        on_events: List[str] = [],
        on_crons: list = [],
        version: str = "",
        timeout: str = "60m",
        schedule_timeout: str = "15m",
        sticky: 'StickyStrategy' = None,
    ) -> Type['WorkflowT']:  # sourcery skip: default-mutable-arg
        """
        Binds the workflow component to the workflow
        """
        self.build_workflow(
            obj = obj,
            name = name,
            on_events = on_events,
            on_crons = on_crons,
            version = version,
            timeout = timeout,
            schedule_timeout = schedule_timeout,
            sticky = sticky,
        )
        return self.c.workflow_meta_class(obj.name, obj.__bases__, dict(obj.__dict__))


    def chain_workflow(
        self,
        ref: 'WorkflowT',
        with_meta: Optional[bool] = True,
        with_links: Optional[bool] = True,
        **kwargs
    ) -> 'WorkflowT':
        """
        Chains the workflow
        """
        for lname, lcls in ref.linked_workflows.items():
            if lcls.name in self.bound_workflows: continue
            self.autologger.info(f'Chaining Workflow: [{lname}] {lcls.name} (With Meta: {with_meta}, With Links: {with_links})', colored = True, prefix = 'Workflow')
            l_wkflow_cls = self.c.workflow_meta_class(lcls.name, lcls.__bases__, dict(lcls.__dict__)) if with_meta else lcls
            l_wkflow: 'WorkflowT' = l_wkflow_cls()
            l_wkflow.on_start(hatchet = self, **kwargs)
            ref.linked_workflows[lname] = l_wkflow
            self.bound_workflows[lcls.name] = l_wkflow
            if lcls.linked_workflows and with_links:
                self.chain_workflow(lcls, with_meta = with_meta, with_links = with_links, **kwargs)
        return ref
    
    def gather_workflow(
        self,
        ref: 'WorkflowT',
        with_meta: Optional[bool] = True,
        with_links: Optional[bool] = True,
        **kwargs
    ) -> 'WorkflowT':
        """
        Gathers the workflow
        """
        if ref.name in self.bound_workflows: return ref
        ref_cls = self.c.workflow_meta_class(ref.name, ref.__bases__, dict(ref.__dict__)) if with_meta else ref
        wkflow: 'WorkflowT' = ref_cls()
        wkflow.on_start(hatchet = self, **kwargs)
        if ref.linked_workflows and with_links:
            self.chain_workflow(wkflow, with_meta = with_meta, with_links = with_links, **kwargs)
        self.bound_workflows[ref.name] = wkflow
        return wkflow

    def gather_workflows(
        self,
        workflow_names: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        **kwargs,
    ) -> Dict[str, 'WorkflowT']:
        """
        Gathers the workflows for the registered workflows
        """
        workflow_names = workflow_names or self.registered_workflows.keys()
        include = include or []
        exclude = exclude or []
        # workflows: List['WorkflowT'] = []
        for workflow_name in workflow_names:
            if workflow_name in self.registered_workflows and (
                not include or workflow_name in include
            ) and (
                not exclude or workflow_name not in exclude
            ):
                if workflow_name in self.bound_workflows: continue
                ref = self.registered_workflows[workflow_name]
                self.gather_workflow(ref, **kwargs)
        return self.bound_workflows
    
    def gather_workers(
        self,
        workflow_names: Optional[List[str]] = None,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        **kwargs,
    ) -> Dict[str, 'Worker']:
        """
        Gathers the workers for the registered workflows
        """
        workflows = self.gather_workflows(workflow_names = workflow_names, include = include, exclude = exclude, **kwargs)
        workers: Dict[str, 'Worker'] = {}
        for workflow_name, workflow in workflows.items():
            self.autologger.info(f'Gathering Worker: {workflow_name}: {workflow}', prefix = f'Workflow ({self.instance})', colored = True)
            worker = self.worker(workflow_name)
            worker.register_workflow(workflow)
            workers[workflow_name] = worker
            self.workers[workflow_name] = worker
        return workers
    

    """
    Compile Method
    """

    def compile_workflow(
        self,
        ref: Type['WorkflowT'],
        **kwargs
    ) -> 'WorkflowT': 
        """
        Compiles the workflow
        """
        if ref.name in self.compiled_workflows: return ref
        # self.autologger.info(f'Compiling Workflow: {ref.name}: {ref}', prefix = 'Workflow', colored = True)
        ref = self.c.workflow_meta_class(ref.name, ref.__bases__, dict(ref.__dict__))()
        ref.on_start(hatchet = self, **kwargs)
        self.compiled_workflows[ref.name] = ref
        return ref
    
    def compile_workflows(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        extra_workflows: Optional[Dict[str, 'WorkflowT']] = None,
        include_crontasks: Optional[bool] = None,
        only_crontasks: Optional[bool] = None,
        **kwargs
    ) -> Dict[str, 'WorkflowT']:
        """
        Compiles the workflows
        """
        all_workflows: Dict[str, 'WorkflowT'] = {}
        workflows: Dict[str, Type['WorkflowT']] = self.c.gather_workflows(include = include, exclude = exclude, only_crontasks = only_crontasks, include_crontasks = include_crontasks, **kwargs)
        if extra_workflows: workflows.update(extra_workflows)
        for workflow_name, wkflow in workflows.items():
            wkflow = wkflow.register(self)
            wkflow = self.compile_workflow(wkflow, **kwargs)
            all_workflows[workflow_name] = wkflow
        # Link them together
        for workflow_name in all_workflows:
            all_workflows[workflow_name].linked_workflows = all_workflows
        return all_workflows


    def delete_all_workflows(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
    ):
        """
        Deletes all workflows
        """
        include = include or []
        exclude = exclude or []
        deleted = []
        workflows = self.client.rest.workflow_list()
        for workflow in workflows.rows:
            workflow_id = workflow.metadata.id
            workflow_name = workflow.name
            if (
                include
                and all(name not in workflow_name for name in include)
            ): continue
            if (
                exclude
                and any(name in workflow_name for name in exclude)
            ): continue
            if exclude and workflow_name in exclude: continue
            self.autologger.info(f'Deleting Workflow: {workflow_name}', prefix = workflow_id, colored = True)
            self.client.rest.workflow_api.workflow_delete(workflow_id)
            deleted.append(workflow_name)
        self.autologger.info(f'Deleted {len(deleted)}/{len(workflows.rows)} Workflows', colored = True)
        return deleted

    async def retrieve_workflow_run(
        self,
        request_id: str,
        blocking: bool = None,
        raise_error: Optional[bool] = False,
        **kwargs,
    ) -> Optional['WorkflowRun']:
        """
        Retrieves the workflow run
        """
        try:
            workflow_ref = self.client.rest.workflow_run_get(request_id)
        except Exception as e:
            if not raise_error: return None
            self.autologger.error(f"Error Retrieving Workflow Run {request_id}: {e}")
            raise e
        if blocking and workflow_ref.status in {
            WorkflowRunStatus.QUEUED,
            WorkflowRunStatus.RUNNING,
            WorkflowRunStatus.PENDING,
        }:
            await self.client.admin.get_workflow_run(request_id).result()
            workflow_ref = self.client.rest.workflow_run_get(request_id)
        return workflow_ref
    
    async def determine_if_workflow_is_active(
        self,
        request_id: str,
        raise_error: Optional[bool] = False,
        **kwargs,
    ) -> bool:
        """
        Determines if the workflow is active
        """
        try:
            workflow_ref = self.client.rest.workflow_run_get(request_id)
        except Exception as e:
            if not raise_error: return False
            self.autologger.error(f"Error Retrieving Workflow Run {request_id}: {e}")
            raise e
        return workflow_ref.status in {
            WorkflowRunStatus.QUEUED,
            WorkflowRunStatus.RUNNING,
            WorkflowRunStatus.PENDING,
        }

    async def abort_workflow_run(
        self,
        request_id: str,
        raise_error: Optional[bool] = False,
        **kwargs,
    ) -> Optional['WorkflowRun']:
        """
        Aborts the workflow run
        """
        try:
            workflow_ref = self.client.rest.workflow_run_get(request_id)
        except Exception as e:
            if not raise_error: return None
            self.autologger.error(f"Error Retrieving Workflow Run {request_id}: {e}")
            raise e
        try:
            cancel_request = WorkflowRunsCancelRequest(
                workflowRunIds = [workflow_ref.metadata.id],
            )
            resp = self.client.rest.workflow_run_api.workflow_run_cancel(
                tenant = self.client.config.tenant_id,
                workflow_runs_cancel_request = cancel_request
            )
            self.autologger.info(f"Canceled Workflow Run {request_id}: {resp}", colored = True)
            workflow_ref = self.client.rest.workflow_run_get(request_id)

        except Exception as e:
            self.autologger.error(f"Error Cancelling Workflow Run {request_id}: {e}")
            if raise_error: raise e
        return workflow_ref


    async def poll_workflow_run(
        self,
        request_id: str,
        **kwargs,
    ) -> Tuple[bool, Union[Dict[str, Any], Any], 'WorkflowRun']:
        """
        Polls the workflow run
        """
        workflow = self.client.admin.get_workflow_run(request_id)
        listener = workflow.stream()
        async for event in listener:
            if event.type == StepRunEventType.STEP_RUN_EVENT_TYPE_COMPLETED:
                return True, event.payload, self.client.rest.workflow_run_get(request_id)
            
            if event.type in [StepRunEventType.STEP_RUN_EVENT_TYPE_FAILED, StepRunEventType.STEP_RUN_EVENT_TYPE_CANCELLED]:
                return False, event.payload, self.client.rest.workflow_run_get(request_id)




from __future__ import annotations

import typing as t
from .client import Client, DEFAULT_RETRIES, DEFAULT_TIMEOUT

if t.TYPE_CHECKING:
    from niquests import Response, AsyncResponse
    from niquests.sessions import HttpAuthenticationType, HttpMethodType, QueryParameterType, BodyType, HeadersType, CookiesType, MultiPartFilesType, MultiPartFilesAltType, TimeoutType, ProxyType, HookType, TLSVerifyType, TLSClientCertType
    from niquests.sessions import PreparedRequest
    from niquests._async import RetryType
    from niquests._typing import AsyncHttpAuthenticationType, AsyncBodyType
    from niquests.structures import QuicSharedCache

_QUIC_CACHE: t.Optional['QuicSharedCache'] = None

def get_shared_quic_cache() -> 'QuicSharedCache':
    """
    Returns the shared quic cache
    """
    global _QUIC_CACHE
    if _QUIC_CACHE is None:
        from niquests.api import _SHARED_QUIC_CACHE
        _QUIC_CACHE = _SHARED_QUIC_CACHE
    return _QUIC_CACHE


@t.overload
def request(
    method: 'HttpMethodType',
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    data: 'BodyType' | None = ...,
    json: t.Any | None = ...,
    headers: 'HeadersType' | None = None,
    cookies: 'CookiesType' | None = None,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    retries: 'RetryType' = ...,
) -> 'Response':
    """Constructs and sends a :class:`Request <Request>`.

    :param method: method for the new :class:`Request` object: ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the :class:`Request`.
    :param files: (optional) Dictionary of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
        ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
        or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content_type'`` is a string
        defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
        to add for the file.
    :param auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How many seconds to wait for the server to send data
        before giving up, as a float, or a :ref:`(connect timeout, read
        timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Boolean. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection. Defaults to ``True``.
    :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``.
            It is also possible to put the certificates (directly) in a string or bytes.
    :param stream: (optional) if ``False``, the response content will be immediately downloaded.
    :param cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    :param hooks: (optional) Register functions that should be called at very specific moment in the request lifecycle.
    :param retries: (optional) If integer, determine the number of retry in case of a timeout or connection error.
            Otherwise, for fine gained retry, use directly a ``Retry`` instance from urllib3.
    :return: :class:`Response <Response>` object

    Usage::

      >>> from lzl.api import aioreq
      >>> req = aioreq.request('GET', 'https://httpbin.org/get')
      >>> req
      <Response HTTP/2 [200]>
    """
    ...

    
@t.overload
def request(
    method: 'HttpMethodType',
    url: str,
    *,
    retries: 'RetryType' = DEFAULT_RETRIES,
    **kwargs: t.Any,
) -> 'Response':
    # By using the 'with' statement we are sure the session is closed, thus we
    # avoid leaving sockets open which can trigger a ResourceWarning in some
    # cases, and look like a memory leak in others.
    with Client(
        quic_cache_layer = get_shared_quic_cache(),
        retries=retries
    ) as session:
        return session.request(
            method=method,
            url=url,
            **kwargs,
        )
    

@t.overload
async def arequest(
    method: 'HttpMethodType',
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    data: 'BodyType' | 'AsyncBodyType' | None = ...,
    json: t.Any | None = ...,
    headers: 'HeadersType' | None = None,
    cookies: 'CookiesType' | None = None,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | 'AsyncHttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: t.Literal[False] = ...,
    cert: 'TLSClientCertType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    retries: 'RetryType' = ...,
) -> 'Response':
    """Constructs and sends a :class:`Request <Request>`.

    :param method: method for the new :class:`Request` object: ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the :class:`Request`.
    :param files: (optional) Dictionary of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
        ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
        or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content_type'`` is a string
        defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
        to add for the file.
    :param auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How many seconds to wait for the server to send data
        before giving up, as a float, or a :ref:`(connect timeout, read
        timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Boolean. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection. Defaults to ``True``.
    :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``.
            It is also possible to put the certificates (directly) in a string or bytes.
    :param stream: (optional) if ``False``, the response content will be immediately downloaded.
    :param cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    :param hooks: (optional) Register functions that should be called at very specific moment in the request lifecycle.
    :param retries: (optional) If integer, determine the number of retry in case of a timeout or connection error.
            Otherwise, for fine gained retry, use directly a ``Retry`` instance from urllib3.
    :return: :class:`Response <Response>` object

    Usage::

      >>> from lzl.api import aioreq
      >>> req = await aioreq.arequest('GET', 'https://httpbin.org/get')
      >>> req
      <Response HTTP/2 [200]>
    """
    ...


@t.overload
async def arequest(
    method: 'HttpMethodType',
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    data: 'BodyType' | 'AsyncBodyType' | None = ...,
    json: t.Any | None = ...,
    headers: 'HeadersType' | None = None,
    cookies: 'CookiesType' | None = None,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | 'AsyncHttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: t.Literal[True] = ...,
    cert: 'TLSClientCertType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    retries: 'RetryType' = ...,
) -> 'AsyncResponse':
    """Constructs and sends a :class:`Request <Request>`.

    :param method: method for the new :class:`Request` object: ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary, list of tuples or bytes to send
        in the query string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) A JSON serializable Python object to send in the body of the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the :class:`Request`.
    :param files: (optional) Dictionary of ``'name': file-like-objects`` (or ``{'name': file-tuple}``) for multipart encoding upload.
        ``file-tuple`` can be a 2-tuple ``('filename', fileobj)``, 3-tuple ``('filename', fileobj, 'content_type')``
        or a 4-tuple ``('filename', fileobj, 'content_type', custom_headers)``, where ``'content_type'`` is a string
        defining the content type of the given file and ``custom_headers`` a dict-like object containing additional headers
        to add for the file.
    :param auth: (optional) Auth tuple to enable Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How many seconds to wait for the server to send data
        before giving up, as a float, or a :ref:`(connect timeout, read
        timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Boolean. Enable/disable GET/OPTIONS/POST/PUT/PATCH/DELETE/HEAD redirection. Defaults to ``True``.
    :param proxies: (optional) Dictionary mapping protocol to the URL of the proxy.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
            the server's TLS certificate, or a string, in which case it must be a path
            to a CA bundle to use. Defaults to ``True``.
            It is also possible to put the certificates (directly) in a string or bytes.
    :param stream: (optional) if ``False``, the response content will be immediately downloaded.
    :param cert: (optional) if String, path to ssl client cert file (.pem). If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    :param hooks: (optional) Register functions that should be called at very specific moment in the request lifecycle.
    :param retries: (optional) If integer, determine the number of retry in case of a timeout or connection error.
            Otherwise, for fine gained retry, use directly a ``Retry`` instance from urllib3.
    :return: :class:`AsyncResponse <AsyncResponse>` object

    Usage::

      >>> from lzl.api import aioreq
      >>> req = await aioreq.arequest('GET', 'https://httpbin.org/get', stream = True)
      >>> req
      <AsyncResponse HTTP/2 [200]>
    """
    ...

async def arequest(
    method: 'HttpMethodType',
    url: str,
    *,
    retries: 'RetryType' = DEFAULT_RETRIES,
    **kwargs: t.Any,
) -> 'Response' | 'AsyncResponse':
    async with Client(
        quic_cache_layer = get_shared_quic_cache(),
        retries=retries
    ) as session:
        return await session.arequest(
            method=method,
            url=url,
            **kwargs,
        )

"""
Sync API Methods
"""

@t.overload
def get(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'Response':
    r"""Sends a GET request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

def get(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response':
    return request("GET", url, **kwargs)


@t.overload
def options(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = True,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'Response':
    r"""Sends a OPTIONS request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

def options(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response':
    return request("OPTIONS", url, **kwargs)


@t.overload
def head(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = True,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'Response':
    r"""Sends a HEAD request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to False by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

def head(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response':
    return request("HEAD", url, **kwargs)



@t.overload
def post(
    url: str,
    data: 'BodyType' | None = ...,
    json: t.Any | None = ...,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
) -> 'Response':
    r"""Sends a POST request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json to send in the body of the
        :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param files: (optional) Dictionary of ``'filename': file-like-objects``
        for multipart encoding upload.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...


def post(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response':
    return request("POST", url, **kwargs)


@t.overload
def put(
    url: str,
    data: 'BodyType' | None = ...,
    *,
    json: t.Any | None = ...,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
) -> 'Response':
    r"""Sends a PUT request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json to send in the body of the
        :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param files: (optional) Dictionary of ``'filename': file-like-objects``
        for multipart encoding upload.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

def put(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response':
    return request("PUT", url, **kwargs)


@t.overload
def patch(
    url: str,
    data: 'BodyType' | None = ...,
    *,
    json: t.Any | None = ...,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
) -> 'Response':
    r"""Sends a PATCH request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json to send in the body of the
        :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param files: (optional) Dictionary of ``'filename': file-like-objects``
        for multipart encoding upload.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

def patch(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response':
    return request("PATCH", url, **kwargs)


@t.overload
def delete(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'Response':
    r"""Sends a DELETE request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

def delete(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response':
    return request("DELETE", url, **kwargs)

"""
Async API Methods
"""


@t.overload
async def aget(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: t.Literal[False] = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'Response':
    r"""Sends a GET request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...


@t.overload
async def aget(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: t.Literal[True] = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'AsyncResponse':
    r"""Sends a GET request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

async def aget(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response' | 'AsyncResponse':
    return await arequest("GET", url, **kwargs)


@t.overload
async def aoptions(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'Response' | 'AsyncResponse':
    r"""Sends a OPTIONS request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

async def aoptions(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response'| 'AsyncResponse':
    return await arequest("OPTIONS", url, **kwargs)


@t.overload
async def ahead(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'Response' | 'AsyncResponse':
    r"""Sends a HEAD request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to False by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

async def ahead(
    url: str,
    *args,
    **kwargs: t.Any,
) -> 'Response' | 'AsyncResponse':
    return await arequest("HEAD", url, **kwargs)


@t.overload
async def apost(
    url: str,
    data: 'BodyType' | None = ...,
    json: t.Any | None = ...,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: t.Literal[False] = ...,
    cert: 'TLSClientCertType' | None = ...,
) -> 'Response':
    r"""Sends a POST request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json to send in the body of the
        :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param files: (optional) Dictionary of ``'filename': file-like-objects``
        for multipart encoding upload.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

@t.overload
async def apost(
    url: str,
    data: 'BodyType' | None = ...,
    json: t.Any | None = ...,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: t.Literal[True] = ...,
    cert: 'TLSClientCertType' | None = ...,
) -> 'AsyncResponse':
    r"""Sends a POST request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json to send in the body of the
        :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param files: (optional) Dictionary of ``'filename': file-like-objects``
        for multipart encoding upload.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

async def apost(
    url: str,
    *args,
    **kwargs,
) -> 'Response' | 'AsyncResponse':
    return await arequest("POST", url, **kwargs)


@t.overload
async def aput(
    url: str,
    data: 'BodyType' | None = ...,
    *,
    json: t.Any | None = ...,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
) -> 'Response' | 'AsyncResponse':
    r"""Sends a PUT request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json to send in the body of the
        :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param files: (optional) Dictionary of ``'filename': file-like-objects``
        for multipart encoding upload.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

async def aput(
    url: str,
    *args,
    **kwargs,
) -> 'Response' | 'AsyncResponse':
    return await arequest("PUT", url, **kwargs)


@t.overload
async def apatch(
    url: str,
    data: 'BodyType' | None = ...,
    *,
    json: t.Any | None = ...,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    files: 'MultiPartFilesType' | 'MultiPartFilesAltType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
) -> 'Response' | 'AsyncResponse':
    r"""Sends a PATCH request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param data: (optional) Dictionary, list of tuples, bytes, or file-like
        object to send in the body of the :class:`Request`.
    :param json: (optional) json to send in the body of the
        :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param files: (optional) Dictionary of ``'filename': file-like-objects``
        for multipart encoding upload.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...
    
async def apatch(
    url: str,
    *args,
    **kwargs,
) -> 'Response' | 'AsyncResponse':
    return await arequest("PATCH", url, **kwargs)


@t.overload
async def adelete(
    url: str,
    *,
    params: 'QueryParameterType' | None = ...,
    headers: 'HeadersType' | None = ...,
    cookies: 'CookiesType' | None = ...,
    auth: 'HttpAuthenticationType' | None = ...,
    timeout: 'TimeoutType' | None = ...,
    allow_redirects: bool = ...,
    proxies: 'ProxyType' | None = ...,
    hooks: 'HookType[PreparedRequest | Response]' | None = ...,
    verify: 'TLSVerifyType' = ...,
    stream: bool = ...,
    cert: 'TLSClientCertType' | None = ...,
    **kwargs: t.Any,
) -> 'Response' | 'AsyncResponse':
    r"""Sends a DELETE request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param params: (optional) Dictionary or bytes to be sent in the query
        string for the :class:`Request`.
    :param headers: (optional) Dictionary of HTTP Headers to send with the
        :class:`Request`.
    :param cookies: (optional) Dict or CookieJar object to send with the
        :class:`Request`.
    :param auth: (optional) Auth tuple or callable to enable
        Basic/Digest/Custom HTTP Auth.
    :param timeout: (optional) How long to wait for the server to send
        data before giving up, as a float, or a :ref:`(connect timeout,
        read timeout) <timeouts>` tuple.
    :param allow_redirects: (optional) Set to True by default.
    :param proxies: (optional) Dictionary mapping protocol or protocol and
        hostname to the URL of the proxy.
    :param hooks: (optional) Dictionary mapping hook name to one event or
        list of events, event must be callable.
    :param stream: (optional) whether to immediately download the response
        content. Defaults to ``False``.
    :param verify: (optional) Either a boolean, in which case it controls whether we verify
        the server's TLS certificate, or a string, in which case it must be a path
        to a CA bundle to use. Defaults to ``True``. When set to
        ``False``, requests will accept any TLS certificate presented by
        the server, and will ignore hostname mismatches and/or expired
        certificates, which will make your application vulnerable to
        man-in-the-middle (MitM) attacks. Setting verify to ``False``
        may be useful during local development or testing.
        It is also possible to put the certificates (directly) in a string or bytes.
    :param cert: (optional) if String, path to ssl client cert file (.pem).
        If Tuple, ('cert', 'key') pair, or ('cert', 'key', 'key_password').
    """
    ...

async def adelete(
    url: str,
    *args,
    **kwargs,
) -> 'Response' | 'AsyncResponse':
    return await arequest("DELETE", url, **kwargs)


from __future__ import annotations

"""Utility helpers for formatting LazyOps log messages."""

import re
import warnings
import typing as t

warnings.filterwarnings("ignore", category=SyntaxWarning)

from .static import (
    COLORED_MESSAGE_MAP,
    DEFAULT_STATUS_COLORS,
    LOGLEVEL_MAPPING,
    RESET_COLOR,
)

if t.TYPE_CHECKING:
    from dataclasses import _DataclassT
    from pydantic import BaseModel

    Primitives = t.Union[str, int, float, bool, tuple, list, type(None)]
    MsgItem = t.Union[BaseModel, _DataclassT, dict[str, t.Any], Primitives]
else:
    MsgItem = t.Any


def find_and_format_seps(msg: str) -> str:
    """Expand short colour directives (``|a,b|`` → ``|a||b|``)."""

    for sep_match in re.finditer(r"\|\w+,(\w+,*)+\|", msg):
        candidate = sep_match.group()
        if len(candidate) >= 10:
            continue
        msg = msg.replace(candidate, "||".join(candidate.split(",")))
    return msg


def extract_module_name(name: str) -> str:
    """Return the root module component from ``name``."""

    return name.split(".")[0]


def get_prefix_and_suffix(
    colored: bool | None = False,
    level: str | None = None,
) -> tuple[str, str]:
    """Return colour prefixes/suffixes for a message when enabled."""

    prefix = suffix = ""
    if colored:
        if level:
            level = level.lower()
            prefix = DEFAULT_STATUS_COLORS.get(level, "|g|")
        suffix = "|e|"
    return prefix, suffix


def format_item(
    msg: MsgItem,
    max_length: int | None = None,
    colored: bool | None = False,
    level: str | None = None,
    _is_part: bool | None = False,
) -> str:  # sourcery skip: extract-duplicate-method, low-code-quality, split-or-ifs
    """Normalise diverse message inputs into a printable string."""

    if isinstance(msg, str):
        return msg[:max_length] if max_length else msg
    if isinstance(msg, (float, int, bool, type(None))):
        rendered = str(msg)
        return rendered[:max_length] if max_length else rendered
    if isinstance(msg, (list, set)):
        rendered = str(msg) if _is_part else "\n" + "".join(f"- {item}\n" for item in msg)
        return rendered[:max_length] if max_length else rendered

    prefix, suffix = get_prefix_and_suffix(colored=colored, level=level)
    if isinstance(msg, dict):
        rendered = "\n"
        for key, value in msg.items():
            value_text = f"{value}"
            if max_length and len(value_text) > max_length:
                value_text = f"{value_text[:max_length]}..."
            rendered += f"- {prefix}{key}{suffix}: {value_text}\n"
        return rendered.rstrip()

    if isinstance(msg, tuple):
        rendered = "".join(
            f"- {prefix}{key}{suffix}: {format_item(value, max_length=max_length, colored=colored, level=level, _is_part=True)}\n"
            for key, value in zip(msg[0], msg[1])
        )
        return rendered[:max_length] if max_length else rendered

    if hasattr(msg, "is_fsspec"):
        return format_item(repr(msg), max_length=max_length, colored=colored, level=level, _is_part=_is_part)

    if (hasattr(msg, "dict") and hasattr(msg, "Config")) or hasattr(msg, "fields"):
        rendered = f"{prefix}[{msg.__class__.__name__}]{suffix}"
        fields = msg.fields.keys() if hasattr(msg, "fields") else msg.__fields__.keys()
        for field in fields:
            field_str = f"{prefix}{field}{suffix}"
            value_text = f"\n  {field_str}: {getattr(msg, field)!r}"
            if max_length is not None and len(value_text) > max_length:
                value_text = f"{value_text[:max_length]}..."
            rendered += value_text
        return rendered

    if hasattr(msg, "model_dump"):
        return format_item(msg.model_dump(mode="json"), max_length=max_length, colored=colored, level=level, _is_part=_is_part)

    if hasattr(msg, "dict"):
        return format_item(msg.dict(), max_length=max_length, colored=colored, level=level, _is_part=_is_part)

    if hasattr(msg, "json"):
        return format_item(msg.json(), max_length=max_length, colored=colored, level=level, _is_part=_is_part)

    if hasattr(msg, "__dict__"):
        return format_item(msg.__dict__, max_length=max_length, colored=colored, level=level, _is_part=_is_part)

    rendered = str(msg)
    return rendered[:max_length] if max_length else rendered


def format_message(
    message: MsgItem,
    *args: MsgItem,
    prefix: str | None = None,
    max_length: int | None = None,
    level: str | None = None,
    colored: bool | None = False,
    extra: dict[str, t.Any] | None = None,
) -> str:
    """Render a message plus optional extras into a single string."""

    rendered = ""
    if prefix:
        if colored and "|" not in prefix:
            base_color = "|g|"
            if level:
                level = level.lower()
                base_color = DEFAULT_STATUS_COLORS.get(level, "|g|")
            prefix = f"{base_color}{prefix}|e|"
        rendered += f"[{prefix}] "
    rendered += format_item(message, max_length=max_length, colored=colored, level=level)
    if args:
        for arg in args:
            rendered += "\n"
            rendered += format_item(arg, max_length=max_length, colored=colored, level=level)
    if extra:
        extras_rendered = format_item(extra, max_length=max_length, colored=colored, level=level)
        extras_rendered = extras_rendered.lstrip("\n")
        if extras_rendered:
            if rendered:
                rendered += "\n"
            rendered += extras_rendered
    if colored:
        rendered = rendered.replace("<fg", ">|fg")
        rendered = rendered.replace("<", "\\</")
        rendered = find_and_format_seps(rendered)
        for key, value in COLORED_MESSAGE_MAP.items():
            rendered = rendered.replace(key, value)
        rendered = rendered.replace(">|fg", "<fg")
        rendered = rendered.replace("\\</", "\\<")
        rendered += RESET_COLOR
    return rendered


def get_logging_level(level: t.Union[str, int]) -> str:
    """Normalise numeric/string levels into upper-case strings."""

    if isinstance(level, str):
        return level.upper()
    if isinstance(level, int):
        return LOGLEVEL_MAPPING.get(level, "INFO")
    return str(level)


__all__ = [
    "MsgItem",
    "find_and_format_seps",
    "extract_module_name",
    "get_prefix_and_suffix",
    "format_item",
    "format_message",
    "get_logging_level",
]



from __future__ import annotations

import json
import pathlib
from pydantic import Field, field_validator
from lzo.types import AppEnv, eproperty
from .base import BaseOpenAISettings, ApiType, DEFAULT_AZURE_VERSION
from .proxy import BaseProxySettings, OpenAIProxySettings
from typing import Optional, Dict, Union, Any



class AzureOpenAISettings(BaseOpenAISettings):
    """
    The Azure OpenAI Settings
    """

    api_type: Optional[ApiType] = ApiType.azure
    api_version: Optional[str] = DEFAULT_AZURE_VERSION
    api_path: Optional[str] = None

    class Config:
        env_prefix = 'AZURE_OPENAI_'
        case_sensitive = False

    @property
    def is_valid(self) -> bool:
        """
        Returns whether the Azure Settings are Valid
        """
        return self.api_key is not None and (
            self.url is not None or self.api_base is not None
        )



class OpenAISettings(BaseOpenAISettings):
    """
    The OpenAI Settings
    """

    app_env: Optional[AppEnv] = None
    client_configurations: Optional[Dict[str, Dict[str, Any]]] = Field(default_factory = dict)
    auto_loadbalance_clients: Optional[bool] = True
    auto_healthcheck: Optional[bool] = True

    function_cache_enabled: Optional[bool] = True
    
    class Config:
        env_prefix = 'OPENAI_'
        case_sensitive = False


    @eproperty
    def azure(self) -> AzureOpenAISettings:
        """
        Returns the Azure Settings
        """
        return AzureOpenAISettings()
    
    # @eproperty
    # def proxy(self) -> OpenAIProxySettings:
    #     """
    #     Return the Proxy Settings
    #     """
    #     return OpenAIProxySettings()
    
    @property
    def has_valid_azure(self) -> bool:
        """
        Returns whether the Azure Settings are Valid
        """
        return self.azure.is_valid

    def load_client_configurations(
        self,
        path: pathlib.Path,
    ):
        """
        Loads the Client Configurations
        """
        if not path.exists(): return
        text = path.read_text()
        if 'env/' in text:
            from lzo.utils import parse_envvars_from_text
            text, _ = parse_envvars_from_text(text)
        assert path.suffix in {
            ".yaml", ".yml", ".json"
        }, f"The preset file must be a YAML or JSON file: {path}"
        
        data: Dict[str, Union[Dict[str, str], str]] = {}
        if path.suffix == ".json":
            data = json.loads(text)
        else:
            import yaml
            data = yaml.safe_load(text)
        self.client_configurations.update(data)

    def configure(
        self, 
        auto_healthcheck: Optional[bool] = None,
        auto_loadbalance_clients: Optional[bool] = None,
        proxy_app_name: Optional[str] = None,
        proxy_env_name: Optional[str] = None,
        proxy_provider: Optional[str] = None,
        proxy_config: Optional[Union[Dict[str, Any], pathlib.Path]] = None,
        client_configurations: Optional[Union[Dict[str, Dict[str, Any]], pathlib.Path]] = None,
        **kwargs
    ):
        """
        Allows Post-Init Configuration of the OpenAI Settings

        Usage:

        ```python
        >>> settings.configure(
        >>>    api_key = 'sk-...',
        >>>    organization = 'org-...',
        >>>    max_retries = 4,
        >>>    timeout = 60,
        >>> )
        ```

        :param url: The OpenAI API URL              | Env: [`OPENAI_API_URL`]
        :param scheme: The OpenAI API Scheme        | Env: [`OPENAI_API_SCHEME`]
        :param host: The OpenAI API Host            | Env: [`OPENAI_API_HOST`]
        :param port: The OpenAI API Port            | Env: [`OPENAI_API_PORT`]
        :param api_key: The OpenAI API Key          | Env: [`OPENAI_API_KEY`]
        :param api_path: The OpenAI API Path        | Env: [`OPENAI_API_PATH`]
        :param api_type: The OpenAI API Type        | Env: [`OPENAI_API_TYPE`]
        :param api_version: The OpenAI API Version  | Env: [`OPENAI_API_VERSION`]
        :param api_key_path: The API Key Path       | Env: [`OPENAI_API_KEY_PATH`]
        :param organization: Organization           | Env: [`OPENAI_ORGANIZATION`]
        :param proxies: The OpenAI Proxies          | Env: [`OPENAI_PROXIES`]
        :param timeout: Timeout in Seconds          | Env: [`OPENAI_TIMEOUT`]
        :param max_retries: The OpenAI Max Retries  | Env: [`OPENAI_MAX_RETRIES`]
        :param kwargs: Additional Keyword Arguments
        """
        if auto_healthcheck is not None: self.auto_healthcheck = auto_healthcheck
        if auto_loadbalance_clients is not None: self.auto_loadbalance_clients = auto_loadbalance_clients
        if proxy_config:
            if not proxy_provider: 
                if isinstance(proxy_config, pathlib.Path):
                    proxy_provider = self.proxy.proxy_provider
                else:
                    proxy_provider = proxy_config.pop('proxy_provider', self.proxy.proxy_provider)
            
            # if not proxy_provider: raise ValueError("Proxy Provider must be provided")
            if isinstance(proxy_config, pathlib.Path):
                self.proxy.load_proxy_config(proxy_config, proxy_provider)
            else: self.proxy.update_proxy_config(proxy_provider, **proxy_config)
            self.proxy.init_proxy()
            self.azure.proxy = self.proxy
        
        elif proxy_provider: 
            self.proxy.current = proxy_provider
            self.azure.proxy.current = proxy_provider
        if proxy_app_name: 
            self.proxy.proxy_app_name = proxy_app_name
            self.azure.proxy.proxy_app_name = proxy_app_name
        if proxy_env_name: 
            self.proxy.proxy_env_name = proxy_env_name
            self.azure.proxy.proxy_env_name = proxy_env_name
        if client_configurations:
            if isinstance(client_configurations, pathlib.Path):
                self.load_client_configurations(client_configurations)
            else: self.client_configurations.update(client_configurations)

        # Parse apart the azure setting configurations
        az_kwargs, rm_keys = {}, []
        for k, v in kwargs.items():
            if k.startswith('azure_'):
                az_kwargs[k[6:]] = v
                rm_keys.append(k)
            
            elif k.startswith('az_'):
                az_kwargs[k[3:]] = v
                rm_keys.append(k)
        
        # Configure the Azure Settings
        if az_kwargs: self.azure.configure(**az_kwargs)
        for k in rm_keys: kwargs.pop(k, None)
        super().configure(**kwargs)

    
    @field_validator('app_env', mode = 'before')
    def validate_app_env(cls, value: Optional[Any]) -> Any:
        """
        Validates the app environment
        """
        if value is None:
            from lzo.types import get_app_env
            return get_app_env('openai')
        return AppEnv.from_env(value) if isinstance(value, str) else value

    @eproperty
    def in_k8s(self) -> bool:
        """
        Returns whether the app is running in kubernetes
        """
        from lzo.utils.system import is_in_kubernetes
        return is_in_kubernetes()
    
    @property
    def is_local_env(self) -> bool:
        """
        Returns whether the environment is development
        """
        return self.app_env in [AppEnv.DEVELOPMENT, AppEnv.LOCAL] and not self.in_k8s
    
    @property
    def is_production_env(self) -> bool:
        """
        Returns whether the environment is production
        """
        return self.app_env == AppEnv.PRODUCTION and self.in_k8s

    @property
    def is_development_env(self) -> bool:
        """
        Returns whether the environment is development
        """
        return self.app_env in [AppEnv.DEVELOPMENT, AppEnv.LOCAL, AppEnv.CICD]




from __future__ import annotations

"""
Temporal Object Registry: Main
"""

import copy
import typing as t
from lzo.registry.base import MRegistry, combine_parts
from lzl.types import eproperty
from lzl.pool import ThreadPool

from temporalio import workflow
from temporalio import activity
from .base import RegistryItem, RegistryIndex
from .methods import configure_workflow, configure_activity, configure_dispatch
from ..utils import logger

if t.TYPE_CHECKING:
    from lzl.io.persistence import TemporaryData
    from temporalio.types import ClassType, CallableType
    from ..client import TemporalClient
    from ..configs import TemporalSettings
    from ..mixins import (
        BaseTemporalMixin, TemporalWorkflowMixin, TemporalActivityMixin, TemporalDispatchMixin,
        TemporalMixinT, TemporalWorkflowT, TemporalActivityT, TemporalDispatchT,
        MixinKinds
    )
    

    
class TemporalRegistry(MRegistry['TemporalMixinT']):
    """
    Temporal Registry
    """
    workflow: RegistryIndex = RegistryIndex()
    activity: RegistryIndex = RegistryIndex()
    dispatch: RegistryIndex = RegistryIndex()


    # Consolidate
    # search_index: t.Dict[str, str] = {}
    # type_index: t.Dict[str, 'TemporalActivityT' | 'TemporalWorkflowT'] = {}
    clients: t.Dict[str, 'TemporalClient'] = {}
    client_get_hooks: t.List[t.Callable[[], t.Awaitable['TemporalClient']]] = []
    client_init_hooks: t.List[t.Callable[[], 'TemporalClient']] = []

    @eproperty
    def client(self) -> t.Optional['TemporalClient']:
        """
        Returns the Temporal Client
        """
        return self._extra.get('client')

    def add_client_hook(
        self, 
        func: t.Callable[[], 'TemporalClient' |  t.Awaitable['TemporalClient']], 
        mode: t.Optional[t.Literal['get', 'init']] = None,
    ):
        """
        Adds a client hook
        """
        # logger.info(f'Adding Client Hook: {func}')
        if mode is None:  mode = 'get' if ThreadPool.is_coro(func) else 'init'
        if mode == 'get': self.client_get_hooks.append(func)
        elif mode == 'init': self.client_init_hooks.append(func)
    
    def run_client_init_hooks(self, client: 'TemporalClient'):
        """
        [Init] Runs the client hooks
        """
        for n, hook in enumerate(self.client_init_hooks):
            if isinstance(hook, str):
                from lzl.load import lazy_import
                hook = lazy_import(hook)
                self.client_init_hooks[n] = hook
            if ThreadPool.is_coro(hook):
                ThreadPool.create_background_task(hook, client)
            else:
                hook(client)
    
    async def run_client_get_hooks(self, client: 'TemporalClient'):
        """
        [Get] Runs the client hooks
        """
        for n, hook in enumerate(self.client_get_hooks):
            if isinstance(hook, str):
                from lzl.load import lazy_import
                hook = lazy_import(hook)
                self.client_get_hooks[n] = hook
            if not ThreadPool.is_coro(hook):
                hook(client)
            else:
                await hook(client)
        
    def register_client(self, client: 'TemporalClient'):
        """
        Registers the Temporal Client
        """
        self.clients[client.namespace] = client
        if not self.client: self.client = client
        self.run_client_init_hooks(client)

    async def get_client(
        self,
        host: t.Optional[str] = None,
        api_key: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        converter: t.Optional[str] = None,
        **kwargs,
    ) -> 'TemporalClient':  # sourcery skip: inline-immediately-returned-variable
        """
        Returns the Temporal Client
        """
        # if namespace and namespace in self.clients: return self.clients[namespace]
        # we'll register null ones too
        if namespace in self.clients: return self.clients[namespace]
        new = await self.config.get_client(
            host = host,
            api_key = api_key,
            namespace = namespace,
            converter = converter,
            **kwargs,
        )
        # do set all the extra stuff
        # ..
        await self.run_client_get_hooks(new)
        return new

    
    @eproperty
    def config(self) -> 'TemporalSettings':
        """
        Returns the Temporal Settings
        """
        from ..configs import get_temporal_settings
        return get_temporal_settings()

    @property
    def tmpdata(self) -> 'TemporaryData':
        """
        Returns the temporary data
        """
        return self.config.tmpdata

    def _register_initialized(self, key: str, value: 'TemporalMixinT') -> None:
        """
        Registers an initialized Temporal Object in the registry
        """
        self.init_registry[key] = value
        if value.mixin_kind == 'workflow': self.workflow.names.add(key)
        elif value.mixin_kind == 'activity': self.activity.names.add(key)
        elif value.mixin_kind == 'dispatch': self.dispatch.names.add(key)
        if value.namespace and value.namespace in self.clients:
            value.client = self.clients[value.namespace]
    

    def _register(self, key: str, value: 'TemporalMixinT') -> None:
        """
        Registers a Temporal Object in the registry
        """
        # self._import_wf_configs_()
        super()._register(key, value)
        if value.mixin_kind == 'workflow': self.workflow.names.add(key)
        elif value.mixin_kind == 'activity': self.activity.names.add(key)
        elif value.mixin_kind == 'dispatch': self.dispatch.names.add(key)

    def register_activity_step(self, func: t.Callable[..., t.Any], **kwargs) -> None:
        """
        Registers an activity step
        """
        pass

    def compile_activity(self, obj: 'TemporalActivityT') -> None:
        """
        Compiles the Temporal Activity
        """
        pass

    def register_workflow_step(self, func: t.Callable[..., t.Any], **kwargs) -> None:
        """
        Registers a workflow step
        """
        pass

    def compile_workflow(self, obj: 'TemporalWorkflowT') -> None:
        """
        Compiles the Temporal Workflow
        """
        pass

    def compile_dispatch(self, obj: 'TemporalDispatchT') -> None:
        """
        Compiles the Temporal Dispatch
        """
        pass

    """
    Main Mixin Registry Methods
    """

    def get_registry_item(self, obj: 'TemporalMixinT', is_type: t.Optional[bool] = None) -> RegistryItem:
        """
        Creates the registry item
        """
        return RegistryItem.build(obj, is_type = is_type)

    """
    Workflow Methods
    """

    def configure_workflow(
        self,
        obj:  t.Type['TemporalWorkflowT'],
        **kwargs,
    ):
        # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
        """
        Configures the workflow
        """
        configure_workflow(obj, **kwargs)


    def register_initialized_workflow(
        self,
        obj: 'TemporalWorkflowT',
        **kwargs,
    ) -> None:  # sourcery skip: class-extract-method
        """
        Registers an initialized workflow
        """
        obj._on_init_hook_()
        if hasattr(obj, '_rxtra'):
            return self._register_initialized(obj._rxtra['registry_name'], obj)
        i = self.get_registry_item(obj, is_type = False)
        if i.registry_name in self.init_registry:
            raise ValueError(f'Workflow {i.registry_name} is already registered')
        return self._register_initialized(i.registry_name, obj)
    
    def register_workflow(
        self,
        obj: t.Union[t.Type['TemporalWorkflowT'], 'TemporalWorkflowT'],
        defn: t.Optional[bool] = None,
        **kwargs,
    ) -> None:
        # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
        """
        Registers the workflow with the registry
        """
        if not isinstance(obj, type): return self.register_initialized_workflow(obj, **kwargs)
        i = self.get_registry_item(obj, is_type = True)
        if i.registry_name in self.mregistry:
            if self.tmpdata.has_logged(f'tmprl:register:workflow:exists:{i.registry_name}'): return
            logger.warning(f'Workflow {i.registry_name} already registered with `{i.registry_name}`', prefix = 'Temporal')
            return
        i.register_obj(obj)
        self.configure_workflow(obj, **kwargs)
        self.compile_workflow(obj)
        self.workflow.index_obj(obj)
        if not defn:
            workflow._Definition._apply_to_class(
                obj,
                workflow_name = None if obj.dynamic else obj.display_name,
                sandboxed = obj.sandboxed,
                failure_exception_types = obj.failure_exception_types or [],
            )
        self[f'workflow.{[i.registry_name]}'] = obj
        if not self.tmpdata.has_logged(f'tmprl:register:workflow:success:{i.registry_name}'):
            extra = f' ({self.client.namespace})' if self.client else ''
            logger.info(f'Registered Workflow: `|g|{i.registry_name}|e|` from `|g|{i.module_source}|e|`{extra}', colored = True, prefix = 'Temporal')

    """
    Activity Methods
    """

    def configure_activity(
        self,
        obj:  t.Type['TemporalActivityT'],
        **kwargs,
    ):
        # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
        """
        Configures the activity
        """
        # init doesnt matter here
        configure_activity(obj, **kwargs)
    
    def register_initialized_activity(
        self,
        obj: 'TemporalActivityT',
        **kwargs,
    ) -> None:
        """
        Registers an initialized activity
        """
        obj._on_init_hook_()
        if hasattr(obj, '_rxtra'):
            return self._register_initialized(obj._rxtra['registry_name'], obj)
        i = self.get_registry_item(obj, is_type = False)
        if i.registry_name in self.init_registry:
            raise ValueError(f'Activity {i.registry_name} is already registered')
        return self._register_initialized(i.registry_name, obj)
    
    def register_activity(
        self,
        obj: t.Union[t.Type['TemporalActivityT'], 'TemporalActivityT'],
        **kwargs,
    ) -> None:
        # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
        """
        Registers the activity with the registry
        """
        if not isinstance(obj, type): return self.register_initialized_activity(obj, **kwargs)
        i = self.get_registry_item(obj, is_type = True)
        if i.registry_name in self.mregistry:
            if self.tmpdata.has_logged(f'tmprl:register:activity:exists:{i.registry_name}'): return
            logger.warning(f'Activity {i.registry_name} already registered with `{i.registry_name}`', prefix = 'Temporal')
            return
        i.register_obj(obj)
        self.configure_activity(obj, **kwargs)
        self.compile_activity(obj)
        self.activity.index_obj(obj)
        self[f'activity.{i.registry_name}'] = obj
        if not self.tmpdata.has_logged(f'tmprl:register:activity:success:{i.registry_name}'):
            extra = f' ({self.client.namespace})' if self.client else ''
            logger.info(f'Registered Activity: `|g|{i.registry_name}|e|` from `|g|{i.module_source}|e|`{extra}', colored = True, prefix = 'Temporal')

    """
    Dispatch Methods
    """

    def configure_dispatch(
        self,
        obj: t.Type['TemporalDispatchT'],
        **kwargs,
    ):
        # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
        """
        Configures the dispatch
        """
        configure_dispatch(obj, **kwargs)

    
    def register_initialized_dispatch(
        self,
        obj: 'TemporalDispatchT',
        **kwargs,
    ) -> None:
        """
        Registers an initialized Dispatch
        """
        obj._on_init_hook_()
        if hasattr(obj, '_rxtra'):
            return self._register_initialized(obj._rxtra['registry_name'], obj)
        i = self.get_registry_item(obj, is_type = False)
        if i.registry_name in self.init_registry:
            raise ValueError(f'Dispatch {i.registry_name} is already registered')
        return self._register_initialized(i.registry_name, obj)
    
    def register_dispatch(
        self,
        obj: t.Union[t.Type['TemporalDispatchT'], 'TemporalDispatchT'],
        **kwargs,
    ) -> None:
        # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
        """
        Registers the dispatch with the registry
        """
        if not isinstance(obj, type): return self.register_initialized_dispatch(obj, **kwargs)
        i = self.get_registry_item(obj, is_type = True)
        if i.registry_name in self.mregistry:
            if self.tmpdata.has_logged(f'tmprl:register:dispatch:exists:{i.registry_name}'): return
            logger.warning(f'Dispatch {i.registry_name} already registered with `{i.registry_name}`', prefix = 'Temporal')
            return
        i.register_obj(obj)
        self.configure_dispatch(obj, **kwargs)
        self.compile_dispatch(obj)
        self.dispatch.index_obj(obj)
        self[f'dispatch.{i.registry_name}'] = obj
        if not self.tmpdata.has_logged(f'tmprl:register:dispatch:success:{i.registry_name}'):
            extra = f' ({self.client.namespace})' if self.client else ''
            logger.info(f'Registered Dispatch: `|g|{i.registry_name}|e|` from `|g|{i.module_source}|e|`{extra}', colored = True, prefix = 'Temporal')

    def _get_index_(self, kind: t.Literal['workflow', 'activity', 'dispatch']) -> RegistryIndex:
        """
        Returns the index
        """
        return getattr(self, kind)

    def search_for_obj(
        self,
        module: t.Optional[str] = None,
        submodule: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        workspace: t.Optional[str] = None,
        name: t.Optional[str] = None,
        display_name: t.Optional[str] = None,
        kind: t.Optional['MixinKinds'] = 'workflow',
        **kwargs,
    ) -> t.Optional[str]:
        """
        Searches for the Temporal Object
        """
        idx = self._get_index_(kind)
        return idx.search_for_obj(module = module, submodule = submodule, namespace = namespace, workspace = workspace, name = name, display_name = display_name, **kwargs)
    
    def register_mixin(
        self,
        obj: t.Union[t.Type['TemporalMixinT'], 'TemporalMixinT'],
        **kwargs,
    ) -> None:
        # sourcery skip: extract-method, inline-immediately-returned-variable, move-assign-in-block
        """
        Registers the Temporal Object with the Registry
        """
        if obj.mixin_kind == 'workflow':
            self.register_workflow(obj, **kwargs)
        elif obj.mixin_kind == 'activity':
            self.register_activity(obj, **kwargs)
        elif obj.mixin_kind == 'dispatch':
            self.register_dispatch(obj, **kwargs)
        
        # else:
        # 
        #     raise ValueError(f'Invalid / Unsupported Mixin Kind: {obj.mixin_kind}')
    
    @t.overload
    def get_workflow(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Literal[True] = ...,
    ) -> t.Type['TemporalWorkflowT'] | t.Type['TemporalWorkflowMixin']:
        """
        Gets a Temporal Workflow Class
        """
        ...
    
    @t.overload
    def get_workflow(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Literal[False] = ...,
    ) -> 'TemporalWorkflowT' | 'TemporalWorkflowMixin':
        """
        Gets a Temporal Workflow Object
        """
        ...

    def get_workflow(
        self,
        registry_name: t.Optional[str] = None,
        module: t.Optional[str] = None,
        submodule: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        workspace: t.Optional[str] = None,
        name: t.Optional[str] = None,
        display_name: t.Optional[str] = None,
        as_type: t.Optional[bool] = None,
    ) -> t.Type['TemporalWorkflowT'] | 'TemporalWorkflowT':
        """
        Gets a Temporal Workflow
        """
        if not registry_name and not module and not namespace and not name and not display_name:
            raise ValueError('Either `registry_name`, `module`, `namespace`, `name`, or `display_name` must be provided')
        if registry_name and registry_name in self.workflow.search_index:
            registry_name = self.workflow.search_index[registry_name]
            return self.workflow.type_index[registry_name] if as_type else self.get(f'workflow.{registry_name}')
        registry_name = self.search_for_obj(
            module = module,
            submodule = submodule,
            namespace = namespace,
            workspace = workspace,
            name = name or registry_name,
            display_name = display_name,
            kind = 'workflow',
        )
        if not registry_name: return None
        if as_type: return self.workflow.type_index[registry_name]
        return self.get(f'workflow.{registry_name}')

    @t.overload
    def get_activity(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Literal[True] = ...,
    ) -> t.Type['TemporalActivityT'] | t.Type['TemporalActivityMixin']:
        """
        Gets a Temporal Activity Class
        """
        ...
    
    @t.overload
    def get_activity(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Literal[False] = ...,
    ) -> 'TemporalActivityT' | 'TemporalActivityMixin':
        """
        Gets a Temporal Activity Object
        """
        ...

    def get_activity(
        self,
        registry_name: t.Optional[str] = None,
        module: t.Optional[str] = None,
        submodule: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        workspace: t.Optional[str] = None,
        name: t.Optional[str] = None,
        display_name: t.Optional[str] = None,
        as_type: t.Optional[bool] = None,
    ) -> t.Optional['TemporalActivityT']:
        """
        Gets a Temporal Activity
        """
        if not registry_name and not module and not workspace and not namespace and not name and not display_name:
            raise ValueError('Either `registry_name`, `module`, `workspace`, `namespace`, `name`, or `display_name` must be provided')
        if registry_name and registry_name in self.activity.search_index:
            registry_name = self.activity.search_index[registry_name]
            return self.activity.type_index[registry_name] if as_type else self.get(f'activity.{registry_name}')
        registry_name = self.search_for_obj(
            module = module,
            submodule = submodule,
            namespace = namespace,
            workspace = workspace,
            name = name or registry_name,
            display_name = display_name,
            kind = 'activity',
        )
        if not registry_name: return None
        if as_type: return self.activity.type_index[registry_name]
        return self.get(f'activity.{registry_name}')

    @t.overload
    def get_dispatch(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Literal[True] = ...,
    ) -> t.Type['TemporalDispatchT'] | t.Type['TemporalDispatchMixin']:
        """
        Gets a Temporal Dispatch Class
        """
        ...
    
    @t.overload
    def get_dispatch(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Literal[False] = ...,
    ) -> 'TemporalDispatchT' | 'TemporalDispatchMixin':
        """
        Gets a Temporal Dispatch Object
        """
        ...

    def get_dispatch(
        self,
        registry_name: t.Optional[str] = None,
        module: t.Optional[str] = None,
        submodule: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        workspace: t.Optional[str] = None,
        name: t.Optional[str] = None,
        display_name: t.Optional[str] = None,
        as_type: t.Optional[bool] = None,
    ) -> t.Optional['TemporalDispatchT']:
        """
        Gets a Temporal Dispatch
        """
        if not registry_name and not module and not workspace and not namespace and not name and not display_name:
            raise ValueError('Either `registry_name`, `module`, `workspace`, `namespace`, `name`, or `display_name` must be provided')
        if registry_name and registry_name in self.dispatch.search_index:
            registry_name = self.dispatch.search_index[registry_name]
            return self.dispatch.type_index[registry_name] if as_type else self.get(f'dispatch.{registry_name}')
        registry_name = self.search_for_obj(
            module = module,
            submodule = submodule,
            namespace = namespace,
            workspace = workspace,
            name = name or registry_name,
            display_name = display_name,
            kind = 'dispatch',
        )
        if not registry_name: return None
        if as_type: return self.dispatch.type_index[registry_name]
        return self.get(f'dispatch.{registry_name}')

    @t.overload
    def get_ref(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Optional[bool] = ...,
        kind: t.Literal['workflow'] = ...,
    ) -> 'TemporalWorkflowT' | 'TemporalWorkflowMixin' | t.Type['TemporalWorkflowT'] | t.Type['TemporalWorkflowMixin']:
        """
        Gets a Temporal Workflow Class or Object
        """
        ...

    @t.overload
    def get_ref(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Optional[bool] = ...,
        kind: t.Literal['activity'] = ...,
    ) -> 'TemporalActivityT' | 'TemporalActivityMixin' | t.Type['TemporalActivityT'] | t.Type['TemporalActivityMixin']:
        """
        Gets a Temporal Activity Class or Object
        """
        ...

    
    @t.overload
    def get_ref(
        self,
        registry_name: t.Optional[str] = ...,
        module: t.Optional[str] = ...,
        submodule: t.Optional[str] = ...,
        namespace: t.Optional[str] = ...,
        workspace: t.Optional[str] = ...,
        name: t.Optional[str] = ...,
        display_name: t.Optional[str] = ...,
        as_type: t.Optional[bool] = ...,
        kind: t.Literal['dispatch'] = ...,
    ) -> 'TemporalDispatchT' | 'TemporalDispatchMixin' | t.Type['TemporalDispatchT'] | t.Type['TemporalDispatchMixin']:
        """
        Gets a Temporal Dispatch Class or Object
        """
        ...

    def get_ref(
        self,
        registry_name: t.Optional[str] = None,
        module: t.Optional[str] = None,
        submodule: t.Optional[str] = None,
        namespace: t.Optional[str] = None,
        workspace: t.Optional[str] = None,
        name: t.Optional[str] = None,
        display_name: t.Optional[str] = None,
        as_type: t.Optional[bool] = None,
        kind: t.Optional['MixinKinds'] = 'workflow',
    ) -> t.Union['TemporalWorkflowT', 'TemporalActivityT']:
        """
        Gets a Temporal Object
        """
        if kind == 'workflow':
            func = self.get_workflow
        elif kind == 'activity':
            func = self.get_activity
        elif kind == 'dispatch':
            func = self.get_dispatch
        else:
            raise ValueError(f'Invalid Kind: {kind}')
        # func = self.get_workflow if kind == 'workflow' else self.get_activity
        return func(
            registry_name = registry_name,
            module = module,
            submodule = submodule,
            namespace = namespace,
            workspace = workspace,
            name = name,
            display_name = display_name,
            as_type = as_type,
        )

    
registry: TemporalRegistry = TemporalRegistry('temporal')

@t.overload
def workflow_defn(
    cls: 'TemporalWorkflowT',
    *,
    name: t.Optional[str] = ...,
    sandboxed: t.Optional[bool] = ...,
    dynamic: t.Optional[bool] = ...,
    failure_exception_types: t.Sequence[t.Type[BaseException]] = ...,
) -> 'TemporalWorkflowT':
    """Decorator for workflow classes.

    This must be set on any registered workflow class (it is ignored if on a
    base class).

    Args:
        cls: The class to decorate.
        name: Name to use for the workflow. Defaults to class ``__name__``. This
            cannot be set if dynamic is set.
        sandboxed: Whether the workflow should run in a sandbox. Default is
            true.
        dynamic: If true, this activity will be dynamic. Dynamic workflows have
            to accept a single 'Sequence[RawValue]' parameter. This cannot be
            set to true if name is present.
        failure_exception_types: The types of exceptions that, if a
            workflow-thrown exception extends, will cause the workflow/update to
            fail instead of suspending the workflow via task failure. These are
            applied in addition to ones set on the worker constructor. If
            ``Exception`` is set, it effectively will fail a workflow/update in
            all user exception cases. WARNING: This setting is experimental.
    """
    ...

def workflow_defn(
    cls: t.Optional['ClassType'] = None,
    *,
    name: t.Optional[str] = None,
    sandboxed: t.Optional[bool] = True,
    dynamic: t.Optional[bool] = False,
    failure_exception_types: t.Sequence[t.Type[BaseException]] = [],
) -> t.Callable[['ClassType'], 'ClassType']:
    """Decorator for workflow classes.

    This must be set on any registered workflow class (it is ignored if on a
    base class).

    Args:
        cls: The class to decorate.
        name: Name to use for the workflow. Defaults to class ``__name__``. This
            cannot be set if dynamic is set.
        sandboxed: Whether the workflow should run in a sandbox. Default is
            true.
        dynamic: If true, this activity will be dynamic. Dynamic workflows have
            to accept a single 'Sequence[RawValue]' parameter. This cannot be
            set to true if name is present.
        failure_exception_types: The types of exceptions that, if a
            workflow-thrown exception extends, will cause the workflow/update to
            fail instead of suspending the workflow via task failure. These are
            applied in addition to ones set on the worker constructor. If
            ``Exception`` is set, it effectively will fail a workflow/update in
            all user exception cases. WARNING: This setting is experimental.
    """
    if cls and workflow._Definition.from_class(cls): return cls
    def decorator(cls: 'ClassType' | t.Type['TemporalWorkflowT']) -> 'ClassType':
        nonlocal name, sandboxed, dynamic, failure_exception_types
        registry.register_workflow(cls, defn = True)
        if cls.failure_exception_types: failure_exception_types = cls.failure_exception_types
        if cls.sandboxed is not None: sandboxed = cls.sandboxed
        if cls.dynamic is not None: dynamic = cls.dynamic
        # name = None if dynamic else cls._rxtra['workflow_name']
        name = None if dynamic else cls.display_name
        # This performs validation
        workflow._Definition._apply_to_class(
            cls,
            workflow_name = name,
            sandboxed = sandboxed,
            failure_exception_types = failure_exception_types,
        )
        return cls
    return decorator(cls) if cls is not None else decorator


def activity_defn(
    cls: 'ClassType' | 'TemporalActivityT',
    fn: 'CallableType',
    *,
    name: t.Optional[str] = None,
    no_thread_cancel_exception: bool = False,
    dynamic: bool = False,
):
    """Decorator for activity functions.

    Activities can be async or non-async.

    Args:
        fn: The function to decorate.
        name: Name to use for the activity. Defaults to function ``__name__``.
            This cannot be set if dynamic is set.
        no_thread_cancel_exception: If set to true, an exception will not be
            raised in synchronous, threaded activities upon cancellation.
        dynamic: If true, this activity will be dynamic. Dynamic activities have
            to accept a single 'Sequence[RawValue]' parameter. This cannot be
            set to true if name is present.
    """
    if cls.no_thread_cancel_exception is not None: no_thread_cancel_exception = cls.no_thread_cancel_exception
    if cls.dynamic is not None: dynamic = cls.dynamic
    def decorator(fn: 'CallableType') -> 'CallableType':
        # This performs validation
        activity._Definition._apply_to_callable(
            fn,
            activity_name = None if dynamic else name or fn.__name__,
            no_thread_cancel_exception = no_thread_cancel_exception,
        )
        return fn
    return decorator(fn)



workflow.defn = workflow_defn


from __future__ import annotations

import os
import time
import socket
import pathlib
import tempfile
import contextlib

from typing import Union, Optional, Dict, Any, List, TYPE_CHECKING
from functools import lru_cache

if TYPE_CHECKING:
    from torch import device

# Stateful Variables
_system_host_name: Optional[str] = None
_system_host_ip: Optional[str] = None


@lru_cache()
def get_torch_device_name(mps_enabled: bool = False):
    """
    Returns the torch device name
    """
    with contextlib.suppress(ImportError):
        import torch
        if torch.cuda.is_available(): return 'cuda'
        if mps_enabled and torch.torch.backends.mps.is_available(): return 'mps'
    return 'cpu'

@lru_cache()
def get_torch_device(mps_enabled: bool = False) -> 'device':
    """
    Returns the torch device
    """
    with contextlib.suppress(ImportError):
        import torch
        return torch.device(get_torch_device_name(mps_enabled))
    return 'cpu'

@lru_cache()
def get_cpu_count() -> int:
    """Get the available CPU count for this system.
    Takes the minimum value from the following locations:
    - Total system cpus available on the host.
    - CPU Affinity (if set)
    - Cgroups limit (if set)
    """
    import sys
    count = os.cpu_count()
    # Check CPU affinity if available
    with contextlib.suppress(Exception):
        import psutil
        affinity_count = len(psutil.Process().cpu_affinity())
        if affinity_count > 0:
            count = min(count, affinity_count)
    # Check cgroups if available
    if sys.platform == "linux":
        with contextlib.suppress(Exception):
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_quota_us") as f:
                quota = int(f.read())
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_period_us") as f:
                period = int(f.read())
            cgroups_count = quota // period
            if cgroups_count > 0:
                count = min(count, cgroups_count)

    return count


@lru_cache()
def is_readonly_dir(path: Union[str, pathlib.Path]) -> bool:
    """
    Check if a directory is read-only.
    """
    if isinstance(path, str): path = pathlib.Path(path)
    if not path.is_dir(): path = path.parent
    try:
        path.mkdir(parents = True, exist_ok = True)
        with tempfile.NamedTemporaryFile('w', dir = path.as_posix()) as f:
            f.write('test')
        return False
    except Exception as e:
        return True



@lru_cache()
def get_host_name():
    """
    Wrapper to handle the case where the hostname is not yet available.
    """
    global _system_host_name
    if _system_host_name is None:
        while _system_host_name is None:
            with contextlib.suppress(Exception):
                _system_host_name = socket.gethostname()
            time.sleep(0.2)
    return _system_host_name


@lru_cache()
def get_host_ip():
    """
    Wrapper to handle the case where the hostname is not yet available.
    """
    global _system_host_ip
    if _system_host_ip is None:
        while _system_host_ip is None:
            with contextlib.suppress(Exception):
                _system_host_ip = socket.gethostbyname(get_host_name())
            time.sleep(0.2)
    return _system_host_ip



@lru_cache()
def fetch_resolver_nameserver(
    path: Optional[str] = '/etc/resolv.conf',
) -> Optional[str]:
    """
    Fetches the nameserver from the resolv.conf
    """
    if path is None: return None
    p = pathlib.Path(path)
    if not p.exists(): return None
    lines = p.read_text().splitlines()
    return next(
        (
            line.split(' ')[-1].strip()
            for line in lines
            if line.startswith('nameserver')
        ),
        None,
    )

    
def get_ulimits():
    import resource
    soft_limit, _ = resource.getrlimit(resource.RLIMIT_NOFILE)
    return soft_limit


def set_ulimits(
    max_connections: int = 500,
    verbose: bool = False,
):
    """
    Sets the system ulimits
    to allow for the maximum number of open connections

    - if the current ulimit > max_connections, then it is ignored
    - if it is less, then we set it.
    """
    import resource
    from lzl.logging import logger
    soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
    if soft_limit > max_connections: return
    if hard_limit < max_connections and verbose:
        logger.warning(f"The current hard limit ({hard_limit}) is less than max_connections ({max_connections}).")
    new_hard_limit = max(hard_limit, max_connections)
    if verbose: logger.info(f"Setting new ulimits to ({soft_limit}, {hard_limit}) -> ({max_connections}, {new_hard_limit})")
    resource.setrlimit(resource.RLIMIT_NOFILE, (max_connections + 10, new_hard_limit))
    new_soft, new_hard = resource.getrlimit(resource.RLIMIT_NOFILE)
    if verbose: logger.info(f"New Limits: ({new_soft}, {new_hard})")







from __future__ import annotations

"""
The File Providers
"""
import abc
import typing as t
from lzl.proxied import ProxyObject
from .s3 import (
    AWSFileSystem,
    AWSAccessor,
    MinioFileSystem,
    MinioAccessor,
    S3CFileSystem,
    S3CAccessor,
    R2FileSystem,
    R2Accessor,
)

AccessorLike = t.Union[
    AWSAccessor,
    MinioAccessor,
    S3CAccessor,
    R2Accessor,
]

FileSystemLike = t.Union[
    AWSFileSystem,
    MinioFileSystem,
    S3CFileSystem,
    R2FileSystem,
]

ProviderMapping: t.Dict[str, t.Tuple[t.Type[FileSystemLike], t.Type[AccessorLike]]] = {
    "s3": (AWSFileSystem, AWSAccessor),
    "minio": (MinioFileSystem, MinioAccessor),
    "s3c": (S3CFileSystem, S3CAccessor),
    "r2": (R2FileSystem, R2Accessor),
}

_ProviderAliases = {
    'aws': 's3',
    'mc': 'minio',
    'mio': 'minio',
    's3fs': 's3',
}

class ProviderFileSystemManager(abc.ABC):
    """
    The File System Manager
    """
    def __init__(self, **kwargs):
        self.accessors: t.Dict[str, AccessorLike] = {}
    
    def get_accessor(self, name: str, _reset: t.Optional[bool] = False, **kwargs) -> AccessorLike:
        """
        Gets the accessor
        """
        if name in _ProviderAliases: name = _ProviderAliases[name]
        if not self.accessors.get(name) or _reset:
            if name not in ProviderMapping:
                raise ValueError(f"Invalid Provider: {name}")
            _fs, _accessor = ProviderMapping[name]
            _fs.build_filesystems()
            _accessor.reload_cfs(**kwargs)
            self.accessors[name] = _accessor
        return self.accessors[name]
    
    def get_filesystem(self, name: str, _reset: t.Optional[bool] = False, **kwargs) -> FileSystemLike:
        """
        Gets the filesystem
        """
        if name in _ProviderAliases: name = _ProviderAliases[name]
        if name not in ProviderMapping:
            raise ValueError(f"Invalid Provider: {name}")
        _fs, _ = ProviderMapping[name]
        if _reset: _fs.build_filesystems(**kwargs)
        return _fs


ProviderManager: ProviderFileSystemManager = ProxyObject(ProviderFileSystemManager)


from __future__ import annotations

"""Registry helpers for configuration settings objects."""

import inspect
from pathlib import Path
import typing as t

from .base import MRegistry, combine_parts

if t.TYPE_CHECKING:
    from lzo.types.base import RegisteredSettings

    RSettingT = t.TypeVar('RSettingT', bound='RegisteredSettings')
else:
    RSettingT = t.TypeVar('RSettingT')

__all__ = [
    'get_app_settings',
    'register_app_settings',
    'register_initialized_settings',
    'register_settings',
]

_sregistry: MRegistry['RSettingT'] = MRegistry('settings')


def register_initialized_settings(settings: 'RSettingT') -> None:
    """Cache an already constructed settings object.

    Args:
        settings: Instance of a :class:`RegisteredSettings` subclass.

    Raises:
        ValueError: If another settings instance is already registered for the
            same module/submodule combination.
    """

    if hasattr(settings, '_rxtra'):
        _sregistry._register_initialized(settings._rxtra['registry_name'], settings)
        return
    cls_module = (
        settings._rmodule
        if getattr(settings, '_rmodule', None) is not None
        else settings.__module__.split('.')[0]
    )
    cls_submodule = settings._rsubmodule if getattr(settings, '_rsubmodule', None) is not None else None
    registry_name = combine_parts(cls_module, cls_submodule)
    if registry_name in _sregistry.init_registry:
        raise ValueError(f'Module {registry_name} already has registered settings')
    _sregistry._register_initialized(registry_name, settings)


def register_settings(settings: t.Union[t.Type['RSettingT'], 'RSettingT']) -> None:
    """Register settings for lazy instantiation and discovery.

    Args:
        settings: Settings class or instance to register. Classes are stored for
            lazy creation; instances are cached immediately.
    """

    if not isinstance(settings, type):
        register_initialized_settings(settings)
        return
    cls_name = _sregistry.get_classname(settings, is_type=True)
    cls_module = (
        settings._rmodule
        if getattr(settings, '_rmodule', None) is not None
        else settings.__module__.split('.')[0]
    )
    cls_submodule = settings._rsubmodule
    registry_name = combine_parts(cls_module, cls_submodule)
    if registry_name in _sregistry.mregistry:
        _sregistry.logger.warning(
            f'Settings {registry_name} already registered with {cls_module}'
        )
        return

    module_config_path = Path(inspect.getfile(settings)).parent
    settings._rxtra['module'] = cls_module
    settings._rxtra['submodule'] = cls_submodule
    settings._rxtra['cls_name'] = cls_name
    settings._rxtra['module_config_path'] = module_config_path
    settings._rxtra['registry_name'] = registry_name
    settings._rxtra['registered'] = True

    if '__main__' not in cls_module:
        p = module_config_path
        m_path, iters = None, 0
        while p.name != cls_module and iters < 4:
            p = p.parent
            iters += 1
            if p.name == cls_module:
                m_path = p
                break
        if m_path is not None:
            settings._rxtra['module_path'] = m_path

    _sregistry[cls_module] = settings


def register_app_settings(
    module: str,
    settings_path: str,
    submodule: t.Optional[str] = None,
) -> None:
    """Register a lazily importable settings object.

    Args:
        module: Top-level package name used for lookup.
        settings_path: Dotted path to the settings class.
        submodule: Optional nested namespace for the settings class.
    """

    registry_name = combine_parts(module, submodule)
    if registry_name in _sregistry.mregistry:
        return
    _sregistry.uninit_registry[registry_name] = settings_path


def get_app_settings(
    module: str,
    submodule: t.Optional[str] = None,
    **kwargs: t.Any,
) -> 'RegisteredSettings':
    """Fetch settings for ``module`` (and optional ``submodule``).

    Args:
        module: Top-level package name used during registration.
        submodule: Optional namespace component for nested settings.
        **kwargs: Forwarded to the ``RegisteredSettings`` constructor when
            instantiation is required.

    Returns:
        Lazily constructed settings instance, cached for future retrievals.
    """

    registry_name = combine_parts(module, submodule)
    return _sregistry.get(registry_name, **kwargs)



from __future__ import annotations


from lzl.logging import logger
from starlette.types import ASGIApp, Message, Receive, Scope, Send
from .types import errors
from starlette.responses import PlainTextResponse
from starlette.authentication import (
    AuthCredentials,
)
from starlette.requests import HTTPConnection
from lzl.ext.fast.types.responses import PrettyJSONResponse
from typing import Optional, Dict, Any, List, Union, Tuple, Callable, Type, Awaitable, TYPE_CHECKING

if TYPE_CHECKING:
    from fastapi import Request, Response
    from .config import OAuth2Config
    from .types.user import OAuth2User, UnauthenticatedUser, TokenT
    from .types.token import OAuth2Token
    from .backend import OAuth2Backend



class OAuth2Middleware:
    """
    The OAuth2 Middleware
    """

    def __init__(
        self,
        app: ASGIApp,
        config: 'OAuth2Config',
        backend_class: Type['OAuth2Backend'],
        on_error: Optional[
            Callable[[HTTPConnection, errors.FastAPIException], Response]
        ] = None,
        **kwargs,
    ):
        """
        The OAuth2 Middleware
        """

        self.app = app
        self.config = config
        self.mtg = self.config.get_manager()
        self.backend_class = backend_class
        self.on_error: Callable[
            [HTTPConnection, errors.FastAPIException], Response
        ] = (on_error if on_error is not None else self.default_on_error)

        excluded_url_paths = [
            '/healthz',
            '/health',
            '/metrics',
            '/openapi.json',
        ]
        if self.config.disabled_paths:
            excluded_url_paths += [
                p for p in self.config.disabled_paths if p not in excluded_url_paths
            ]
        if self.mtg.routes.get('callback'):
            excluded_url_paths.append(self.mtg.routes['callback'])
            excluded_url_paths.extend(
                self.mtg.routes['callback'] + "/" + provider
                for provider in self.mtg.providers
            )
        if self.mtg.routes.get('start'):
            excluded_url_paths.append(self.mtg.routes['start'])
            excluded_url_paths.extend(
                self.mtg.routes['start'] + "/" + provider
                for provider in self.mtg.providers
            )
        self.excluded_url_paths = set(excluded_url_paths)
        self.backend = self.backend_class(self.config)
        self.mtg.oauth2_middleware = self
        if self.mtg.excluded_paths:
            self.excluded_url_paths.update(self.mtg.excluded_paths)

    @property
    def middleware_handlers(self) -> List[Callable[..., Awaitable['Scope']]]:
        """
        Returns the middleware handlers
        """
        return self.mtg.custom_middleware_handlers

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        The OAuth2 Middleware
        """
        if scope["type"] not in ["http", "websocket"]:
            await self.app(scope, receive, send)
            return
        if scope['path'] in self.excluded_url_paths:
            await self.app(scope, receive, send)
            return
        conn = HTTPConnection(scope)
        try:
            auth_result = await self.backend.authenticate(conn)
        except errors.OAuth2Exception as exc:
            response = self.on_error(conn, exc)
            if scope["type"] == "websocket":
                await send({"type": "websocket.close", "code": 1000})
            else:
                await response(scope, receive, send)
            return

        if auth_result is None:
            auth_result = AuthCredentials(), self.mtg.unauth_user_class()
        scope["auth"], scope["user"] = auth_result
        if self.middleware_handlers:
            for handler in self.middleware_handlers:
                scope = await handler(scope)
        await self.app(scope, receive, send)

    
    @staticmethod
    def default_on_error(conn: HTTPConnection, exc: Union[Exception, errors.FastAPIException]) -> PrettyJSONResponse:
        """
        Returns the default error response
        """
        if hasattr(exc, 'json_data'):
            return PrettyJSONResponse(content = exc.json_data(), status_code=exc.status_code, headers = exc.headers)
        return PlainTextResponse(str(exc), status_code=400)



from __future__ import annotations

import json
from lzl.io import File
from pathlib import Path
from pydantic_settings import BaseSettings
from typing import Optional, Dict, Any, List, TYPE_CHECKING
from .types import SlackContext


lib_path = Path(__file__).parent
context_path = lib_path.joinpath('slack_context.json')

if TYPE_CHECKING:
    from lzl.io import File

class SlackSettings(BaseSettings):
    """
    Slack Settings
    """

    webhook_url: Optional[str] = None
    bot_token: Optional[str] = None
    signing_secret: Optional[str] = None
    app_token: Optional[str] = None

    default_user: Optional[str] = None
    default_channel: Optional[str] = None
    cache_context: Optional[bool] = True
    context_file: Optional[File] = context_path

    class Config:
        env_prefix = 'SLACK_'
        case_sensitive = False
        extra = 'allow'

    def load_context(self, ctx_file: Optional[File] = None) -> SlackContext:
        """
        Loads the Slack Context
        """
        if ctx_file is not None:
            # assert ctx_file.exists(), f"Context File does not exist: {ctx_file}"
            self.context_file = ctx_file

        if not self.context_file.exists(): return SlackContext()
        return SlackContext(
            **json.loads(self.context_file.read_text())
        )
        # if not context_path.exists(): return SlackContext()
        # return SlackContext(
        #     **json.loads(context_path.read_text())
        # )
    
    def save_context(self, context: SlackContext) -> None:
        """
        Saves the Slack Context
        """
        self.context_file.write_text(json.dumps(context.model_dump(), indent = 4))
        # context_path.write_text(json.dumps(context.model_dump(), indent = 4))



from __future__ import annotations

import typing as t
from lzl.types import eproperty
from pydantic import PrivateAttr, model_validator, field_validator, Field, Json, AliasChoices
from pydantic_settings import BaseSettings, SettingsConfigDict
from lzl.io import File


class KeycloakSettings(BaseSettings):

    url: t.Optional[str] = Field(None, validation_alias=AliasChoices('KEYCLOAK_URL', 'KEYCLOAK_ADMIN_URL'))
    realm: t.Optional[str] = Field(None, validation_alias=AliasChoices('KEYCLOAK_REALM', 'KEYCLOAK_ADMIN_REALM'))
    client_id: t.Optional[str] = Field('admin-cli', validation_alias=AliasChoices('KEYCLOAK_CLIENT_ID', 'KEYCLOAK_ADMIN_CLIENT_ID'))
    username: t.Optional[str] = Field(None, validation_alias=AliasChoices('KEYCLOAK_USERNAME', 'KEYCLOAK_ADMIN_USERNAME'))
    password: t.Optional[str] = Field(None, validation_alias=AliasChoices('KEYCLOAK_PASSWORD', 'KEYCLOAK_ADMIN_PASSWORD'))
    refresh_realms_interval: int = Field(120.0, validation_alias=AliasChoices('KEYCLOAK_REFRESH_REALMS_INTERVAL', 'KEYCLOAK_ADMIN_REFRESH_REALMS_INTERVAL'))
    _extra: t.Dict[str, t.Any] = PrivateAttr(default_factory= dict)

    model_config = SettingsConfigDict(
        env_prefix='',
        case_sensitive=False,
        env_file='.env',
        extra='allow',
    )
    
    @eproperty
    def keycloak_bucket(self) -> File:
        """
        Get the path to the Keycloak bucket.
        """
        return self._extra.get('keycloak_bucket')
    
    @eproperty
    def instance_id(self) -> str:
        """
        Get the instance ID.
        """
        return self._extra.get('instance_id', 'default-instance')




from __future__ import annotations

"""Shared state and registration helpers for LazyOps logging."""

import functools
import logging
import threading
import typing as t

from .utils import extract_module_name

if t.TYPE_CHECKING:
    from logging import LogRecord

_registered_logger_modules: set[str] = set()
_module_name_mapping: dict[str, str] = {}
_module_name_mapping_relative: dict[str, bool] = {}

_is_global_muted: bool | None = None


def set_global_muted(muted: bool) -> None:
    """Toggle a process-wide muted flag checked by LazyOps loggers."""

    global _is_global_muted
    _is_global_muted = muted


def is_global_muted() -> bool:
    """Return ``True`` when the global mute flag is enabled."""

    return bool(_is_global_muted)


def register_logger_module(module: str) -> None:
    """Record ``module`` as having an explicit LazyOps logger configured."""

    _registered_logger_modules.add(extract_module_name(module))


@functools.lru_cache(maxsize=1000)
def is_registered_logger_module(name: str) -> bool:
    """Return ``True`` when ``name`` (or its root) has been registered."""

    module_name = extract_module_name(name)
    return module_name in _registered_logger_modules


def register_module_name(module_name: str, module: str, is_relative: bool = False) -> None:
    """Store a friendly mapping used when patching log records."""

    _module_name_mapping[module_name] = module
    _module_name_mapping_relative[module_name] = is_relative


def run_record_patching_hook(record: t.Union[logging.LogRecord, dict[str, t.Any]]) -> t.Union[logging.LogRecord, dict[str, t.Any]]:
    """Attach resolved module names to Loguru records when available."""

    if record["name"] in _module_name_mapping:
        record["extra"]["module_name"] = _module_name_mapping[record["name"]]
    elif any(record["name"].startswith(module) for module in _module_name_mapping):
        matching_module = next(
            (module for module in _module_name_mapping if record["name"].startswith(module)),
            None,
        )
        if matching_module and _module_name_mapping_relative.get(matching_module, False):
            module = record["name"].replace(matching_module, _module_name_mapping[matching_module]).strip()
            register_module_name(record["name"], module, is_relative=False)
            record["extra"]["module_name"] = _module_name_mapping[record["name"]]
    return record


def add_api_log_filters(
    modules: t.Optional[t.Union[t.Sequence[str], str]] = ("gunicorn", "uvicorn"),
    routes: t.Optional[t.Union[t.Sequence[str], str]] = ("/healthz",),
    status_codes: t.Optional[t.Union[t.Sequence[int], int]] = None,
    verbose: bool = False,
) -> None:  # sourcery skip: default-mutable-arg
    """Attach filters to common HTTP server loggers to hide noisy endpoints."""

    modules_list = [modules] if isinstance(modules, str) else list(modules or [])
    routes_list = [routes] if isinstance(routes, str) else list(routes or [])
    status_list = [status_codes] if isinstance(status_codes, int) else list(status_codes or [])

    def filter_api_record(record: logging.LogRecord) -> bool:
        if routes_list:
            for route in routes_list:
                if route in record.args:
                    return False
        if status_list:
            for sc in status_list:
                if sc in record.args:
                    return False
        return True

    for module in modules_list:
        target = module
        if module == "gunicorn":
            target = "gunicorn.glogging.Logger"
        elif module == "uvicorn":
            target = "uvicorn.logging.Logger"
        api_logger = logging.getLogger(target)
        from .main import default_logger

        if verbose:
            default_logger.info(
                "Adding API filters to %s for routes=%s status_codes=%s",
                target,
                routes_list or None,
                status_list or None,
            )
        api_logger.addFilter(filter_api_record)


__all__ = [
    "set_global_muted",
    "is_global_muted",
    "register_logger_module",
    "is_registered_logger_module",
    "register_module_name",
    "run_record_patching_hook",
    "add_api_log_filters",
]
